/**
 * ----------------------------------------------------------------------------------
 *
 * Copyright (C) 2011 runlevel3 GmbH Berlin
 *
 * ----------------------------------------------------------------------------------
 *
 * LICENSE:
 *
 * This file is part of USB-AD-Interface.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * ----------------------------------------------------------------------------------
 *
 * AVR Studio 4.18 / avr-gcc version 4.3.3
 *
 * ----------------------------------------------------------------------------------
 *
 * @copyright 2011 runlevel3 GmbH
 * @license   http://www.gnu.org/licenses/gpl.txt GNU General Public License v3
 * @link      http://www.runlevel3.de runlevel3
 **/


#include <avr/io.h>
#include <util/delay.h>

#include <ad7797.h>




void AD7797_SPI_start(void)
{
	//slave select ad7797 -> high
	AD7797_SPI_PORT |= _BV(SPI_BIT_SS);

	//set MOSI, SCK and SS output, MISO input
	AD7797_SPI_DDR |= _BV(SPI_BIT_SS) | _BV(SPI_BIT_SCK) | _BV(SPI_BIT_MOSI);
	AD7797_SPI_DDR &= ~_BV(SPI_BIT_MISO);

	//Enable SPI, Master, set clock rate fck/16
	SPCR = _BV(SPE) | _BV(MSTR) | _BV(SPR0);
}


void AD7797_SPI_stop(void)
{
	//slave select ad7797 -> high
	AD7797_SPI_PORT |= _BV(SPI_BIT_SS);

	//disable SPI, Master
	SPCR = 0;
}


unsigned long AD7797_SPI_transfer(unsigned char command, unsigned char numbytes, unsigned long data)
{
	unsigned char cnt;
	unsigned long rv = 0;

	//slave select ad7797 -> low
	AD7797_SPI_PORT &= ~_BV(SPI_BIT_SS);
	
	//Start transmission, send command
	SPDR = command;
	//Wait for transmission complete
	while(!(SPSR & (1<<SPIF))) {};

	for (cnt=numbytes; cnt>0; cnt--)
	{
		SPDR = (data >> (8 * (cnt - 1))) & 0xff;
		//Wait for transmission complete
		while(!(SPSR & (1<<SPIF))) {};
		rv = (rv << 8) | SPDR;
	}

	//slave select ad7797 -> high
	AD7797_SPI_PORT |= _BV(SPI_BIT_SS);
	
	return rv;
}


unsigned char AD7797_busy(unsigned char wait)
{
	//select status register for next read
	AD7797_SPI_transfer(AD7797_RS_COM | AD7797_COM_READ, AD7797_RS_COM_SIZE, 0);
	while (AD7797_SPI_transfer(AD7797_RS_COM | AD7797_COM_READ, AD7797_RS_COM_SIZE, 0) & AD7797_STAT_RDY) {
		if (!wait) return 1;
	}
		
	return 0;
}


unsigned char AD7797_init(void)
{
	AD7797_SPI_start();

	//reset entire interface with 40 bits 1s
	AD7797_SPI_transfer(0xff, 4, 0xffffffff);

	//two dummy reads, returns always 0s
	AD7797_SPI_transfer(AD7797_RS_ID | AD7797_COM_READ, AD7797_RS_ID_SIZE, 0xffffffff);
	AD7797_SPI_transfer(AD7797_RS_ID | AD7797_COM_READ, AD7797_RS_ID_SIZE, 0xffffffff);
	
	//return chip-ID
	return AD7797_SPI_transfer(AD7797_RS_ID | AD7797_COM_READ, AD7797_RS_ID_SIZE, 0);
}


unsigned long AD7797_gain_systemcalibration(unsigned char wait)
{
	//select mode register and set mode selfcalibration
	AD7797_SPI_transfer(AD7797_RS_MODE, AD7797_RS_MODE_SIZE, AD7797_MODE_FSCAL_SYS | AD7797_MODE_DEFAULT);

	AD7797_busy(wait);

	//read and return offset
	return AD7797_offset_read();
}


unsigned long AD7797_offset_systemcalibration(unsigned char wait)
{
	//select mode register and set mode selfcalibration
	AD7797_SPI_transfer(AD7797_RS_MODE, AD7797_RS_MODE_SIZE, AD7797_MODE_OFFSET_SYS | AD7797_MODE_DEFAULT);

	AD7797_busy(wait);

	//read and return offset
	return AD7797_offset_read();
}


unsigned long AD7797_offset_selfcalibration(unsigned char wait)
{
	//select mode register and set mode selfcalibration
	AD7797_SPI_transfer(AD7797_RS_MODE, AD7797_RS_MODE_SIZE, AD7797_MODE_OFFSET_SELF | AD7797_MODE_DEFAULT);

	AD7797_busy(wait);

	//read and return offset
	return AD7797_offset_read();
}


unsigned long AD7797_gain_read(void)
{
	return AD7797_SPI_transfer(AD7797_RS_FS | AD7797_COM_READ, AD7797_RS_FS_SIZE, 0);
}


void AD7797_gain_write(unsigned long gain)
{
	unsigned int mode_save;

	//save mode register
	mode_save =  AD7797_SPI_transfer(AD7797_RS_MODE | AD7797_COM_READ, AD7797_RS_MODE_SIZE, 0);

	//select mode register and set mode idle to write to gain register
	AD7797_SPI_transfer(AD7797_RS_MODE, AD7797_RS_MODE_SIZE, AD7797_MODE_IDLE | AD7797_MODE_DEFAULT);

	//write gain register
	AD7797_SPI_transfer(AD7797_RS_FS, AD7797_RS_FS_SIZE, gain);

	//restore mode register
	AD7797_SPI_transfer(AD7797_RS_MODE, AD7797_RS_MODE_SIZE, mode_save);
}


unsigned long AD7797_offset_read(void)
{
	return AD7797_SPI_transfer(AD7797_RS_OFFSET | AD7797_COM_READ, AD7797_RS_OFFSET_SIZE, 0);
}


void AD7797_offset_write(unsigned long offset)
{
	unsigned int mode_save;

	//save mode register
	mode_save =  AD7797_SPI_transfer(AD7797_RS_MODE | AD7797_COM_READ, AD7797_RS_MODE_SIZE, 0);

	//select mode register and set mode idle to write to gain register
	AD7797_SPI_transfer(AD7797_RS_MODE, AD7797_RS_MODE_SIZE, AD7797_MODE_IDLE | AD7797_MODE_DEFAULT);

	//write offset register
	AD7797_SPI_transfer(AD7797_RS_OFFSET, AD7797_RS_OFFSET_SIZE, offset);

	//restore mode register
	AD7797_SPI_transfer(AD7797_RS_MODE, AD7797_RS_MODE_SIZE, mode_save);
}


unsigned long AD7797_convert(unsigned char chnum, unsigned char uni, unsigned char burnout, unsigned char wait)
{
	unsigned long cfgtmp;

	//select channel, burnout and unipolar
	cfgtmp = AD7797_CFG_1 | (chnum & 7);
	if (uni) cfgtmp |= AD7797_CFG_UNI;
	if (burnout) cfgtmp |= AD7797_CFG_BO;
	AD7797_SPI_transfer(AD7797_RS_CONF, AD7797_RS_CONF_SIZE, cfgtmp);

	//select mode register and set mode convert continuous
	AD7797_SPI_transfer(AD7797_RS_MODE, AD7797_RS_MODE_SIZE, AD7797_MODE_CONTINUOUS | AD7797_MODE_DEFAULT);

	AD7797_busy(wait);

	//command = conversion, dummy data
	return AD7797_SPI_transfer(AD7797_RS_DATA | AD7797_COM_READ, AD7797_RS_DATA_SIZE, 0);
}

unsigned long AD7797_convert_read(void)
{
	//command = read data register
	return AD7797_SPI_transfer(AD7797_RS_DATA | AD7797_COM_READ, AD7797_RS_DATA_SIZE, 0);
}
