/**
 * ----------------------------------------------------------------------------------
 *
 * Copyright (C) 2011 runlevel3 GmbH Berlin
 *
 * ----------------------------------------------------------------------------------
 *
 * LICENSE:
 *
 * This file is part of USB-AD-Interface.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * ----------------------------------------------------------------------------------
 *
 * AVR Studio 4.18 / avr-gcc version 4.3.3
 *
 * ----------------------------------------------------------------------------------
 *
 * @copyright 2011 runlevel3 GmbH
 * @license   http://www.gnu.org/licenses/gpl.txt GNU General Public License v3
 * @link      http://www.runlevel3.de runlevel3
 **/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <avr/pgmspace.h>

#include <util/delay.h>

#include <usbdrv/usbdrv.h>

#include <config.h>
#include <uart-int.h>
#include <timer_x.h>
#include <ad7797.h>
#include <usb-commands.h>


//return datatype
typedef union {
	unsigned long rvl;
	double rvd;
	unsigned char rvc[sizeof(unsigned long)];
} longfloatchar;

// USB report descriptor
PROGMEM char usbHidReportDescriptor[30] = {
    0x06, 0x00, 0xff,              // USAGE_PAGE (Vendor Defined Page 1)
	0x09, 0x01,                    // USAGE (Vendor Usage 1)
    0xa1, 0x01,                    // COLLECTION (Application)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x26, 0xff, 0x00,              //   LOGICAL_MAXIMUM (255)
    0x75, 0x08,                    //   REPORT_SIZE (8)
    0x95, 0x08,                    //   REPORT_COUNT (8)
	0x09, 0x02,                    //   USAGE (Vendor Usage 2)
    0x81, 0x02,                    //   INPUT (Data,Var,Abs)
    0x09, 0x03,                    //   USAGE (Vendor Usage 3)
    0x91, 0x02,                    //   OUTPUT (Data,Var,Abs)
    0x09, 0x04,                    //   USAGE (Vendor Usage 4)
    0xb2, 0x02, 0x01,              //   FEATURE (Data,Var,Abs,Buf)
    0xc0                           // END_COLLECTION
};

//usb transfer variables
unsigned char usbpos, usbleft, usbindex, usbcommand, idlerate;
    
//global config record
extern _config *config;

//auto convert timer [ms]
extern volatile unsigned long convert_timer, offset_timer, gain_timer;


//buffer for statistics
unsigned char stat_ready = 0;
unsigned int stat_index = 0;
unsigned long stat_buff[STAT_LOOPS_MAX];
unsigned long sort_stat_buff[STAT_LOOPS_MAX];
unsigned long mean = 0, median = 0;
double variance = 0.0, std_dev = 0.0;

//global buffer
volatile unsigned long ad_value = 0, offset_new, gain_new;
volatile unsigned char ad_ready = 0, timer_signals = 0, convert_last = 0, convert_new;


void reset_statistics(void)
{
	mean = 0;
	median = 0;
	variance = 0.0;
	std_dev = 0.0;
	stat_index = 0;
	stat_ready = 0;
}

void statistics(void)
{
	unsigned char sortready;
	unsigned int cnt;
	unsigned long long tmpvalue;
	unsigned long tmpmedian;

	//first calculate mean;
	tmpvalue = 0;
	for (cnt = 0; cnt < config->stat_loops; cnt++) {
		tmpvalue += stat_buff[cnt];
		if (cnt % 128 == 0) usbPoll();
	}
	mean = (unsigned long)(tmpvalue / config->stat_loops);
	#ifdef DEBUG
	printf_P(PSTR("mean:         0x%06lX\n"), mean);
	#endif

	usbPoll();

	//calculate variance
	tmpvalue = 0;
	for (cnt = 0; cnt < config->stat_loops; cnt++) {
		tmpvalue += (stat_buff[cnt] - mean) * (stat_buff[cnt] - mean);
		if (cnt % 128 == 0) usbPoll();
	}
	variance = (double)tmpvalue / (double)(config->stat_loops - 1);
	std_dev = sqrt(variance);
	#ifdef DEBUG
	printf_P(PSTR("variance:     %.3f\n"), variance);
	printf_P(PSTR("std_dev:      %.3f\n"), std_dev);
	#endif

	usbPoll();

	for (cnt = 0; cnt < config->stat_loops; cnt++) sort_stat_buff[cnt] = stat_buff[cnt];
	//pseudo quicksort
	do {
		sortready = 1;
		for (cnt = 0; cnt < config->stat_loops - 1; cnt++) {
			if (sort_stat_buff[cnt] > sort_stat_buff[cnt+1]) {
				sortready = 0;
				tmpmedian = sort_stat_buff[cnt];
				sort_stat_buff[cnt] = sort_stat_buff[cnt+1];
				sort_stat_buff[cnt+1] = tmpmedian;
			}
			if (cnt % 128 == 0) usbPoll();
		}

		usbPoll();
	} while (!sortready);

	median = sort_stat_buff[config->stat_loops/2];
	#ifdef DEBUG
	printf_P(PSTR("median:       0x%06lX\n"), median);
	#endif
}

unsigned long save_stat(unsigned long convert_value)
{
	//save convert value in buffer and calculate statistic values
	stat_buff[stat_index++] = convert_value;
	if (stat_index == config->stat_loops) {
		stat_index = 0;
		stat_ready = 1;
	}
	if (stat_ready) statistics();

	return convert_value;
}

unsigned char usbFunctionRead(unsigned char *data, unsigned char len)
{
	unsigned char cnt;
	longfloatchar rv;

    if (!usbleft) return 0;						//transfer complete -> stop
	if(len > usbleft)                			//len is max chunk size
        len = usbleft; 			              	//send an incomplete chunk
    usbleft -= len;

	switch (usbcommand) {
		case USB_CMD_OFFSET:
			rv.rvl = config->offset;
			#ifdef DEBUG
			printf_P(PSTR("offset:       "));
			#endif
			break;
		case USB_CMD_GAIN:
			rv.rvl = config->gain;
			#ifdef DEBUG
			printf_P(PSTR("gain:         "));
			#endif
			break;
		case USB_CMD_OFFSET_SELFCALIBRATION:
			timer_signals |= SIGNAL_OFFSET_SELFCALIBRATION;
			rv.rvl = config->offset;
			#ifdef DEBUG
			printf_P(PSTR("offset_self:  "));
			#endif
			break;
		case USB_CMD_OFFSET_SYSTEMCALIBRATION:
			timer_signals |= SIGNAL_OFFSET_SYSTEMCALIBRATION;
			rv.rvl = config->offset;
			#ifdef DEBUG
			printf_P(PSTR("offset_system:"));
			#endif
			break;
		case USB_CMD_GAIN_SYSTEMCALIBRATION:
			timer_signals |= SIGNAL_GAIN_SYSTEMCALIBRATION;
			rv.rvl = config->gain;
			#ifdef DEBUG
			printf_P(PSTR("gain_system:  "));
			#endif
			break;
		case USB_CMD_CONVERT_AINP_AINM ... 0xFF:
			convert_new = usbcommand;
			#ifdef DEBUG
			printf_P(PSTR("convert_cmd:  0x%02X\n"), usbcommand);
			#endif
			timer_signals |= SIGNAL_CONVERT;
			rv.rvl = ad_value;
			#ifdef DEBUG
			printf_P(PSTR("ad_value:     "));
			#endif
			break;
		case USB_CMD_READ_MEAN:
			rv.rvl = mean;
			#ifdef DEBUG
			printf_P(PSTR("mean:         "));
			#endif
			break;
		case USB_CMD_READ_MEDIAN:
			rv.rvl = median;
			#ifdef DEBUG
			printf_P(PSTR("median:       "));
			#endif
			break;
		case USB_CMD_READ_VARIANCE:
			rv.rvd = variance;
			#ifdef DEBUG
			printf_P(PSTR("variance:     "));
			#endif
			break;
		case USB_CMD_READ_STD_DEV:
			rv.rvd = std_dev;
			#ifdef DEBUG
			printf_P(PSTR("std_dev:      "));
			#endif
			break;
		case USB_CMD_AUTO:
			rv.rvl = config->auto_convert ? config->auto_interval : 0;
			#ifdef DEBUG
			printf_P(PSTR("auto_convert: %lums\n"), rv.rvl);
			#endif
			break;
		case USB_CMD_STAT_LOOPS:
			rv.rvl = config->stat_loops;
			#ifdef DEBUG
			printf_P(PSTR("stat_loops:   %lu\n"), rv.rvl);
			#endif
			break;
		case USB_CMD_STARTUP_OFFSET:
			rv.rvl = config->startup_offset ? config->startup_offset : 0;
			#ifdef DEBUG
			printf_P(PSTR("start_offset: %lu\n"), rv.rvl);
			#endif
			break;
		default:
			return 0;								//return data size
	}
	
	#ifdef DEBUG
	if ((usbcommand == USB_CMD_READ_VARIANCE) || (usbcommand == USB_CMD_READ_STD_DEV)) {
		printf_P(PSTR("%.3f\n"), rv.rvd);					//print value
	} else {
		if (!((usbcommand == USB_CMD_AUTO) || (usbcommand == USB_CMD_STAT_LOOPS))) {
			printf_P(PSTR("0x%06lX\n"), rv.rvl);			//print value
		}
	}
	#endif
	
	memset(data, 0, 8);
	for (cnt=0;cnt<4;cnt++) {
		data[cnt+USB_VALUE_OFFSET] = rv.rvc[3-cnt];
	}

	#ifdef DEBUG1
	printf_P(PSTR("USB data rd:  0x%02X%02X%02X%02X%02X%02X%02X%02X\n"),\
		data[0], data[1], data[2], data[3], data[4], data[5], data[6], data[7]);
	#endif

	return 8;										//return data size
}



unsigned char usbFunctionWrite(unsigned char *data, unsigned char len)
{
	unsigned char cnt;
	unsigned long tmpvalue = 0;

    if (!usbleft) return 1;						//transfer complete -> stop
    if(len > usbleft)                			//len is max chunk size
        len = usbleft; 			              	//send an incomplete chunk
    usbleft -= len;
	
	#ifdef DEBUG1
	printf_P(PSTR("USB data wr:  0x%02X%02X%02X%02X%02X%02X%02X%02X\n"),\
		data[0], data[1], data[2], data[3], data[4], data[5], data[6], data[7]);
	#endif

	switch (data[0]) {
		case USB_CMD_OFFSET:
			for (cnt=0;cnt<4;cnt++) tmpvalue = (tmpvalue << 8) | data[cnt+USB_VALUE_OFFSET];
			offset_new = tmpvalue;
			timer_signals |= SIGNAL_WRITE_OFFSET;
			#ifdef DEBUG
			printf_P(PSTR("write_offset: 0x%06lX\n"), tmpvalue);
			#endif
			return 1;							//transfer complete
		case USB_CMD_GAIN:
			for (cnt=0;cnt<4;cnt++) tmpvalue = (tmpvalue << 8) | data[cnt+USB_VALUE_OFFSET];
			gain_new = tmpvalue;
			timer_signals |= SIGNAL_WRITE_GAIN;
			#ifdef DEBUG
			printf_P(PSTR("write_gain:   0x%06lX\n"), tmpvalue);
			#endif
			return 1;							//transfer complete
		case USB_CMD_WRITE_CONFIG:
			write_config();
			#ifdef DEBUG
			printf_P(PSTR("config_save\n"));
			#endif
			return 1;							//transfer complete
		case USB_CMD_RESET_CONFIG:
			config_defaults();
			write_config();
			reset_statistics();
			gain_new = config->gain;
			offset_new = config->offset;
			timer_signals |= SIGNAL_WRITE_OFFSET | SIGNAL_WRITE_GAIN;
			#ifdef DEBUG
			printf_P(PSTR("config_defaults\n"));
			#endif
			return 1;							//transfer complete
		case USB_CMD_REBOOT:
			wdt_enable(WDTO_30MS);
			#ifdef DEBUG
			printf_P(PSTR("reboot...\n"));
			#endif
			while (1) {};
			return 1;							//transfer complete
		case USB_CMD_AUTO:
			for (cnt=0;cnt<4;cnt++) tmpvalue = (tmpvalue << 8) | data[cnt+USB_VALUE_OFFSET];
			if (!tmpvalue) {
				config->auto_convert = 0;
				convert_timer = 0;
				#ifdef DEBUG
				puts_P(PSTR("auto_convert: off"));
				#endif
			} else {
				config->auto_convert = 1;
				config->auto_mode = convert_last;
				config->auto_interval = tmpvalue;
				//signal start autoconversion for main loop
				convert_new = convert_last;
				convert_last = 0;
			}
			return 1;							//transfer complete
		case USB_CMD_STAT_LOOPS:
			for (cnt=0;cnt<4;cnt++) tmpvalue = (tmpvalue << 8) | data[cnt+USB_VALUE_OFFSET];
			if (tmpvalue > STAT_LOOPS_MAX) tmpvalue = STAT_LOOPS_MAX;
			if (tmpvalue < 2) tmpvalue = 2;
			config->stat_loops = tmpvalue;
			reset_statistics();
			#ifdef DEBUG
			printf_P(PSTR("stat_loops:   %lu\n"), tmpvalue);
			#endif
			return 1;							//transfer complete
		case USB_CMD_WRITE_COMMAND:
			usbcommand = data[7];
			#ifdef DEBUG
			printf_P(PSTR("write_cmd:    0x%02X\n"), data[7]);
			#endif
			return 1;							//transfer complete
		case USB_CMD_STARTUP_OFFSET:
			#ifdef DEBUG
			printf_P(PSTR("start_offset: "));
			#endif
			if (data[7]) {
				#ifdef DEBUG
				puts_P(PSTR("on"));
				#endif
				config->startup_offset = 1;
			} else {
				#ifdef DEBUG
				puts_P(PSTR("off"));
				#endif
				config->startup_offset = 0;
			}
			return 1;							//transfer complete
		default:
			return 1;							//unknown command -> transfer complete
	}
}


usbMsgLen_t usbFunctionSetup(unsigned char data[8])
{
	usbRequest_t *rq = (void *)data;

	#ifdef DEBUG1
	printf_P(PSTR("USB request:\nRequestType:  0x%02X\nRequest:      0x%02X\nValue:        0x%04X\nIndex:        0x%04X\nLength:       0x%04X\n"),\
		rq->bmRequestType, rq->bRequest, rq->wValue.word, rq->wIndex.word, rq->wLength.word);
	#endif    

	if((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS){    // HID class request
		switch (rq->bRequest) {
			case USBRQ_HID_SET_REPORT:
			case USBRQ_HID_GET_REPORT:
		        usbpos = 0;
		        usbleft = rq->wLength.word;  			//size requested
				usbindex = rq->wIndex.word;
	            return USB_NO_MSG;						//use usbFunctionWrite() / usbFunctionRead()
			break;
			case USBRQ_HID_GET_IDLE:
	            usbMsgPtr = &idlerate;
	            return 1;
			break;
			case USBRQ_HID_SET_IDLE:
	            idlerate = rq->wValue.bytes[1];
				return 0;
			break;
		}
    }else{
        //no vendor requests -> debug warning
		puts_P(PSTR("USB VENDOR request! check USB message"));
    }
    return 0;
}





int main(void)
{
	extern unsigned char usbDeviceAddr;
	unsigned char chip_id;
	int cnt;
	
    wdt_enable(WDTO_2S);

	FILE mystdinouterr = FDEV_SETUP_STREAM(USART_putchar, USART_getchar, _FDEV_SETUP_RW);
	stdout = stdin = stderr = &mystdinouterr;

    usbInit();
	// enforce re-enumeration, do this while interrupts are disabled!
    usbDeviceDisconnect();
	// fake USB disconnect for > 250 ms
    for (cnt = 0; cnt < 300; cnt++) {
        wdt_reset();
        _delay_ms(1);
    }
    usbDeviceConnect();

    sei();

    // wait for USB connect
	for (cnt = 0; cnt < 100; cnt++) {
        wdt_reset();
        _delay_ms(1);
    }

	if (USART_init(BAUD19200, DOUBLEBAUD, 1200, 8)) {
		puts_P(PSTR("USART initialized..."));
	}

	printf_P(PSTR("A/D-Chip found ID: 0x%02X: "), chip_id = AD7797_init());
	switch (chip_id) {
		case AD7796_ID:
			puts_P(PSTR("AD7796"));
			break;
		case AD7797_ID:
			puts_P(PSTR("AD7797"));
			break;
		default:
			puts_P(PSTR("unknown"));
	}

	//reset switch -> enable pullup and wait for stable level
	RESET_PORT |= _BV(RESET_BIT);
	_delay_ms(1);
	//pushbutton not pressed, read last settings
	if (RESET_PIN & _BV(RESET_BIT)) {
		printf_P(PSTR("read config..."));
		if (!read_config()) printf_P(PSTR("invalid -> defaults loaded..."));
	} else {
		printf_P(PSTR("reset button pressed -> loading defaults..."));
		read_config();
		//pushbutton pressed, load defaults
		config_defaults();
		write_config();
	}
	puts_P(PSTR("OK"));

	AD7797_offset_write(config->offset);
	AD7797_gain_write(config->gain);
	printf_P(PSTR("mode:         0x%04lX\n"), AD7797_SPI_transfer(AD7797_RS_MODE | AD7797_COM_READ, AD7797_RS_MODE_SIZE, 0));
	printf_P(PSTR("config:       0x%04lX\n"), AD7797_SPI_transfer(AD7797_RS_CONF | AD7797_COM_READ, AD7797_RS_CONF_SIZE, 0));
	printf_P(PSTR("offset:       0x%06lX\ngain:         0x%06lX\n"), config->offset, config->gain);

	if (config->startup_offset) {
		//initiate systemcalibration
		config->offset = AD7797_offset_systemcalibration(1);
		printf_P(PSTR("offset_sys:   0x%06lX\n"), config->offset);
	}

	LED_DDR |= _BV(LED_BIT);

	//start default autoconvert mode and do first conversion
	convert_last = convert_new = AUTO_CONVERT_MODE;
	if (config->auto_convert) convert_last = convert_new = config->auto_mode;
	AD7797_convert(convert_last & 7, (convert_last & 0x40) >> 6, (convert_last & 0x80) >> 7, 0);
	timer_signals |= SIGNAL_CONVERT;

	//timer init _after_ AD7797 init functions
	Timer1_init();

	//main loop -> watchdog reset -> serve usb -> display status
    for(;;) {
        wdt_reset();
        usbPoll();
		//set USB-connect LED
		if (usbDeviceAddr) {
			LED_PORT &= ~_BV(LED_BIT);
		} else {
			LED_PORT |= _BV(LED_BIT);
		}

		//new convert settings -> discard statistic buffer -> send convert command
		if (convert_new != convert_last) {
			convert_last = convert_new;

			//statistics only with identical settings
			reset_statistics();

 			if (config->auto_convert) {
				//start conversion loop after 1ms
				convert_timer = 1;
				printf("auto_convert: %lums start\n", config->auto_interval);
			} else {
				timer_signals |= SIGNAL_CONVERT;
			}
		}

		if (ad_ready) {
			ad_ready = 0;
			if (config->auto_convert) {
				printf_P(PSTR("auto_read:    "));
			} else {
				printf_P(PSTR("convert:      "));
			}
			
			printf_P(PSTR("0x%06lX\n"), ad_value);
			//calculate statistics
			save_stat(ad_value);
		}
    }

    return 0;
}
